"""
 Copyright (c) 2020 Alan Yorinks All rights reserved.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
 Version 3 as published by the Free Software Foundation; either
 or (at your option) any later version.
 This library is distributed in the hope that it will be useful,f
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License for more details.

 You should have received a copy of the GNU AFFERO GENERAL PUBLIC LICENSE
 along with this library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

 DHT support courtesy of Martyn Wheeler
 Based on the DHTNew library - https://github.com/RobTillaart/DHTNew
"""

import sys
import time
import threading
from telemetrix import telemetrix

pinout = {}  # Leave empty
# Dictionary to hold analog pins' raw ADC values.
ADC_results = {
    0: 0,  # pin 0 -> "A0" : 0V
    1: 0,
    2: 0,
    3: 0,
    4: 0,
    5: 0,
    6: 0
}

# Digital pins based on Arduino Digital pin numbering.
# When number is used for digital_ functions, the number is according to digital pins.  (13 = D13 != A13)
output_pins = {
    "d_LED": 13,  # Internal LED on MEGA
}

# When number is used for analog_ functions, the number is according to analog pins ??  (13 = A13 != D13)
# Alternatively use the digital pin numberation instead.  54 = D54 = A0
analog_input_pins = {
    "a_ADC_A0": 0,
    "a_ADC_A1": 1,
    "a_ADC_A2": 2,
    "a_ADC_A3": 3,
    "a_ADC_A4": 4,
    "a_ADC_A5": 5,
    "a_ADC_A6": 6
}

# merge all pins to one dictionary for easier referencing
pinout = analog_input_pins | output_pins

led_state = 1 # for this example

# Mega has 10bits, thus 0 - 1023
def convert_analog(analog_reading: int, Vs: int = 5):
    return (analog_reading * (Vs / 1023))

## ================================================================================================================

# specific class with Telemetrix interface
class Test:
    def __init__(self, com_port=None, arduino_instance_id=1):
        self.board = telemetrix.Telemetrix(com_port, arduino_instance_id)
        self.read_analog_timeout = 3  # in seconds.
        self.cb_flag = 0 # callback filter flag. Only do stuff when requested.

        # Configure pins specific to this example
        self.configure_pins()  # Configure test pin definitions. Assumes pins dictionary is well defined

    # Analog input callback function
    def cb_analog_input_read(self, data):
        '''
        Refreshes every analog input readout. This implementation guarantees, the pin of interest is updated.
        So setting any analog pin mode, can be used to read any analog pin result once updated.
        :param data: required by Telemetrix
        :return: currently nothing
        '''
        if (self.cb_flag == 1): # Filter. Only act when requested. Otherwise callback is empty and terminates quickly.
            print(f'Stored value: Pin Type: {data[0]} Pin: {data[1]} Value: {data[2]}')
            ADC_results[data[1]] = data[2]  # Store the measured value at the corresponding analog pin index
            self.cb_flag = 0  # reset the flag - otherwise no way of stopping the reporting spam.

        ''' message just to show how much callback is piling up and increases calls within the script flow '''
        print(f'Spamming callback from  Pin: {data[1]}')
        # return data  # not needed at the moment as no way of getting the return from callback ?


    def configure_pins(self):
        # If you setting the mode for multiple pins, you may wish to add a short delay after setting each pin.
        # https: // mryslab.github.io / telemetrix / pin_modes /

        # Outputs
        for x in output_pins:  # iterating through outputs to configure them as outputs in Arduino
            self.board.set_pin_mode_digital_output(output_pins[x])
            time.sleep(.02)
        print(f'# Digital output set done')

        # Analog inputs
        # This seems unnecessary to be done at startup. Since during read_analog, this pretty much repeats per pin.
        print("# Setting analog inputs:")
        for x in analog_input_pins:
            # Once pin_mode is set, callbacks are automatically enabled.
            self.board.set_pin_mode_analog_input(analog_input_pins[x], callback=self.cb_analog_input_read)
            self.board.disable_analog_reporting(analog_input_pins[x]) # immediately stop reporting spam
            time.sleep(.02)
        self.board.set_analog_scan_interval(200)
        print(f'# Analog input set done')

    def read_analog(self, a_pin_name: str, delay=0.5):
        '''
        :param a_pin_name: string of a pin name written in the dictionary pinout
        :param delay:   default 0 seconds. Time stalled until reading a measurement.
        :return: ([0]return value, [1]return message)
                 [0] - 0-5V - Specific pin's converted value of raw 10bit ADC value;
                 -1 - cb_flag is not reset during timeout
                 -2 - Reading callback not reached - something wrong with calling the callback;
                 [1] "status message")
        '''
        time.sleep(delay)  # wait for the delay before reading measurement. use >0.1 when making first measurements.
        pin = pinout[a_pin_name]  # get the pin number from dictionary
        self.cb_flag = 1 # set flag to store values in callback

        print(f'# Reading: {a_pin_name}, A{pin}')
        time_now = time.time()  # used to do timeout checks.
        while (self.cb_flag == 1): # as long as flag is set
            if time.time() <= time_now + self.read_analog_timeout:  # implemented some timeout
                # Setting any analog pin causes callback to be activated automatically !
                self.board.set_pin_mode_analog_input(pin, callback=self.cb_analog_input_read)
                time.sleep(0.1)
            else:
                print("analog read timeout")
                return (-1, "analog read timeout")

        if (self.cb_flag == 0):  # check if callback had been reached
            # return converted value
            # print(f'read_analog: {a_pin_name} on {pin}, raw: {ADC_results[pin]}') # dbg
            rslt = convert_analog(ADC_results[pin])
            return (rslt, 'A' + str(pin) + ': ' + str(ADC_results[pin]))
            # result is stored in ADC_results dictionary via callback function.
            # cb_flag is reset also in callback
        else:
            print("analog read callback not reached")
            return (-2, "analog read callback not reached")


    def LED_blink(self, dir_delay: int = 0.5):
        '''
        Call this as a separate Thread !
        :param dir_delay: Time to wait until direction is changed
        :return:
        '''
        global led_state
        time_now = time.time()  # get time when function was called
        while True:  # infinite loop
            if (time.time() > (time_now + dir_delay)):  # When timeout has reached
                time_now = time.time()  # update the start of timer
                self.board.digital_write(pinout["d_LED"], led_state)
                print(f"Blinking LED {led_state}")
                led_state ^= 1
            pass  # else do nothing

# ===============================================================================================================
# ===============================================================================================================

# Create a Telemetrix instance.
# board = telemetrix.Telemetrix()
test = Test()  # Creates an ServOCP test instance
print(f'# Telemetrix instance creation successful\n')
time.sleep(1)

try:

    # background task
    print("# LED Blink")
    # set 'daemon = True' to terminate function when main program exits.
    led = threading.Thread(target=test.LED_blink, args=(0.5,), daemon=True)
    led.start()

    time.sleep(2)

    print("# Read single ADCs")
    print(f'Test acknowledges: {test.read_analog("a_ADC_A0")}') # A0
    time.sleep(1)
    print(f'Test acknowledges: {test.read_analog("a_ADC_A6")}') # A6
    time.sleep(1)
    print(f'Test acknowledges: {test.read_analog("a_ADC_A2")}') # A2
    time.sleep(1)
    print(f'Test acknowledges: {test.read_analog("a_ADC_A3")}') # A3
    # at this point callbacks are very frequent by all analog pins.
    # Although I was interested in the callback at the time of the call, not later.

    time.sleep(2)

    print("# Test script finished")

except KeyboardInterrupt:
    test.board.shutdown()
    sys.exit(0)

finally:
    test.board.shutdown()
    sys.exit(0)


"""
 Copyright (c) 2023 Alan Yorinks All rights reserved.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
 Version 3 as published by the Free Software Foundation; either
 or (at your option) any later version.
 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License for more details.

 You should have received a copy of the GNU AFFERO GENERAL PUBLIC LICENSE
 along with this library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

import sys
import time
from telemetrix import telemetrix

"""
This program reads x, y, and z registers of an ADXL345 using 4 Wire SPI interface
"""

chip_select = 15


def the_callback(data):
    """
    :param data: [pin_type, chip select pin, device read register, x data pair,
                 y data pair, z data pair, time stamp]

    """
    report_type = data[0]
    chip_select_pin = data[1]
    device_read_register = data[2] & 0x3f  # strip off read command bits
    number_bytes_read = data[3]
    x_msb = data[4]
    x_lsb = data[5]
    y_msb = data[6]
    y_lsb = data[7]
    z_msb = data[8]
    z_lsb = data[9]

    x_data = (x_msb << 8) + x_lsb
    y_data = (y_msb << 8) + y_lsb
    z_data = (z_msb << 8) + z_lsb

    # test report type for SPI report
    if report_type == 13:
        print(f'SPI Report:  CS Pin: {chip_select_pin}   SPI Register: '
              f'{device_read_register}   Number Of Bytes Read: {number_bytes_read}    x: '
              f'{x_data}   y: {y_data}   z: {z_data}')
    else:
        print(f'unexpected report type: {report_type}')


def adxl345(my_board):
    """

    :type my_board: object
    """
    # initialize spi mode for chipselect on pin 10
    my_board.set_pin_mode_spi([chip_select])
    time.sleep(.3)

    # set the SPI format
    # spi speed is FPU frequency divided by 4
    # data order is MSB
    # mode is MODE3
    my_board.spi_set_format(4, 1, 3)
    time.sleep(.3)

    # set up power and control register
    my_board.spi_write_blocking(chip_select, [45, 0])
    time.sleep(.3)

    my_board.spi_write_blocking(chip_select, [45, 8])
    time.sleep(.3)

    # set up data format register for 4 wire spi
    my_board.spi_write_blocking(chip_select, [49, 0])
    time.sleep(.3)

    # read 6 bytes from the data register
    # for a multibyte read, we need to OR in a 0x40 into the register value
    while True:
        # read 6 bytes from the data register
        try:
            my_board.spi_read_blocking(chip_select, 50 | 0x40, 6, the_callback)

            time.sleep(.5)

        except (KeyboardInterrupt, RuntimeError):
            my_board.shutdown()
            sys.exit(0)


board = telemetrix.Telemetrix(arduino_wait=4, ip_address="192.168.2.199")
try:
    adxl345(board)
except KeyboardInterrupt:
    board.shutdown()
    sys.exit(0)



stepper_info = {'instance': False, 'is_running': None,
                             'maximum_speed': 1, 'speed': 0, 'acceleration': 0,
                             'distance_to_go_callback': None,
                             'target_position_callback': None,
                             'current_position_callback': None,
                             'is_running_callback': None,
                             'motion_complete_callback': None,
                             'acceleration_callback': None}

stepper_info_list = []

# a list of dictionaries to hold stepper information
for motor in range(8):
    stepper_info_list.append(stepper_info)

stepper_info_list[0]['is_running'] = True

for motor in range(8):
    running = stepper_info_list[motor]['is_running']
    print(f'entry: {motor} value: {running}')



import json
import msgpack
import serial
import sys
import time
from collections import deque

# Sketch uses 3576 bytes (11%) of program storage space. Maximum is 32256 bytes.
# Global variables use 567 bytes (27%) of dynamic memory, leaving 1481 bytes for local variables. Maximum is 2048 bytes.


class JsonDeserialize:
    def __init__(self):
        self.serial_port = serial.Serial('/dev/ttyACM0', 115200,
                                         timeout=1, writeTimeout=0)
        self.start_time = 0
        self.end_time = 0

        print('open')
        count = 0

        length = 0
        while True:
            # try:
            if self.serial_port.inWaiting():
                # the first byte retrieved is the length of the packet
                length = int.from_bytes(self.serial_port.read(), "big")
                packet = self.serial_port.read(length)
                t = msgpack.unpackb(packet, raw=False)
                if t['count'] == 0:
                    self.start_time = time.time()

                if t["count"] == 999:
                    print(f'Elapsed time: {time.time() - self.start_time}')
                    sys.exit(0)
            else:
                try:
                    time.sleep(.000001)
                except KeyboardInterrupt:
                    sys.exit(0)

JsonDeserialize()



import json
import serial
import sys
import time


class JsonDeserialize:
    """
    Deserialize data from ArduinoJson.
    1000 packets will be sent. Show the time for the
    all the packets to be received and deserialized.

    Elapsed time = 8.813957691192627
    
    """

    def __init__(self):
        self.serial_port = serial.Serial('/dev/ttyACM0', 115200,
                                         timeout=1, writeTimeout=0)
        self.serial_port.reset_output_buffer()
        self.serial_port.reset_input_buffer()

        # Time that data started to arrive
        self.start_time = 0
        print('open')

        while True:
            try:
                if self.serial_port.inWaiting():
                    c = self.serial_port.read_until(b'}')
                    try:
                        b = json.loads(c)
                    except (json.decoder.JSONDecodeError, UnicodeDecodeError):
                        continue

                    # Save the start time when the first packet arrives
                    if b['count'] == 0:
                        self.start_time = time.time()

                    # calculate elapsed time when the last packet arrives
                    if b['count'] >= 999:
                        print(f'Elapsed time = {time.time() - self.start_time}')
                        sys.exit(0)
                else:
                    time.sleep(.000001)
                    # continue
            except OSError:
                pass


JsonDeserialize()



import sys
import time

from telemetrix import telemetrix




board = telemetrix.Telemetrix()
ENABLE_PIN =  8
board.set_pin_mode_digital_output(pin_number=ENABLE_PIN)
board.digital_write(pin=ENABLE_PIN,value=0)

def the_callback(data):
    date = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(data[2]))
    print(f'Motor {data[1]} runSpeedToPosition motion completed at: {date}.')


motor = board.set_pin_mode_stepper(interface=1, pin1=2, pin2=5)
board.stepper_set_max_speed(motor, 800)
board.stepper_move_to(motor, 10)
# board.stepper_set_speed(motor, 400)    #can not use  stepper_set_speed two times


motor1 = board.set_pin_mode_stepper(interface=1, pin1=3, pin2=6)
board.stepper_set_max_speed(motor1, 800)
board.stepper_move_to(motor1, 10)
# board.stepper_set_speed(motor1, 400)




# run the motor
board.stepper_run_speed_to_position(motor, completion_callback=the_callback)
board.stepper_run_speed_to_position(motor1, completion_callback=the_callback)

# keep application running
while True:
    try:
        time.sleep(1)
    except KeyboardInterrupt:
        board.shutdown()
        sys.exit(0)





#!/usr/bin/env python3

import sys
import time
from telemetrix import telemetrix
from i2c_lcd_pcf8574 import LcdI2c


"""
This example sets up and control an pcf8574 i2c LCD display.
"""


def pcf8574(my_board):
    lcd = LcdI2c(my_board)
    lcd.begin()
    lcd.clear()
    lcd.backlight()

    lcd.setCursor(5, 1)  # Or setting the cursor in the desired position.
    lcd.print("World!")
    while True:
        time.sleep(1)




board = telemetrix.Telemetrix()
try:
    pcf8574(board)
except KeyboardInterrupt:
    board.shutdown()
    sys.exit(0)



"""
 Copyright (c) 2020 Alan Yorinks All rights reserved.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
 Version 3 as published by the Free Software Foundation; either
 or (at your option) any later version.
 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License for more details.

 You should have received a copy of the GNU AFFERO GENERAL PUBLIC LICENSE
 along with this library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

import asyncio
import sys
import time
from telemetrix import telemetrix

"""
This program continuously monitors an HC-SR04 Ultrasonic Sensor
It reports changes to the distance sensed.
"""
TRIGGER_PIN = 9
ECHO_PIN = 10

# indices into callback data
REPORT_TYPE = 0
TRIG_PIN = 1
DISTANCE = 2
TIME = 3


# A callback function to display the distance
def the_callback(data):
    """
    The callback function to display the change in distance
    :param data: [report_type = PrivateConstants.SONAR_DISTANCE, trigger pin number, distance, timestamp]
    """
    date = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(data[3]))
    print(f'Sonar Report: Trigger Pin: {data[1]} Distance: {data[2]} Time: {date}')


def sonar(my_board, trigger_pin, echo_pin, callback):
    """
    Set the pin mode for a sonar device. Results will appear via the
    callback.

    :param my_board: a telemetrix_aio instance
    :param trigger_pin: Arduino pin number
    :param echo_pin: Arduino pin number
    :param callback: The callback function
    """

    # set the pin mode for the trigger and echo pins
    my_board.set_pin_mode_sonar(trigger_pin, echo_pin, callback)
    # wait forever
    while True:
        try:
            time.sleep(5)
            my_board.sonar_disable()
            time.sleep(5)
            my_board.sonar_enable()

        except KeyboardInterrupt:
            my_board.shutdown()
            sys.exit(0)

# instantiate telemetrix_aio
board = telemetrix.Telemetrix()
try:
    sonar(board, TRIGGER_PIN, ECHO_PIN, the_callback)
    board.shutdown()
except (KeyboardInterrupt, RuntimeError):
    board.shutdown()
    sys.exit(0)



#!/usr/bin/env python3
"""
 Copyright (c) 2020 Alan Yorinks All rights reserved.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
 Version 3 as published by the Free Software Foundation; either
 or (at your option) any later version.
 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License for more details.

 You should have received a copy of the GNU AFFERO GENERAL PUBLIC LICENSE
 along with this library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

import sys
import time
from telemetrix import telemetrix
from i2c_lcd_pcf8574 import LCD_I2C


"""
This example sets up and control an pcf8574 i2c LCD display.
"""
def pcf8574(board):
    lcd = LCD_I2C(board)
    lcd.backlight()
    # time.sleep(1)
    lcd.begin()
    # time.sleep(1)

    # lcd.backlight()
    # time.sleep(1)

    print('printing H')
    lcd.print('Hello World')

    # Flashing the backlight
    # for i in range(5):
    #     lcd.backlight()
    #     time.sleep(1)
    #     lcd.noBacklight()
    #     time.sleep(1)

#    lcd.print("     Hello") # You can make spaces using well... spaces
#    lcd.setCursor(5, 1) # Or setting the cursor in the desired position.
#    lcd.print("World!")
#    lcd.delayMiliseconds(500)
#
#    # Flashing the backlight
#    for i in range(5):
#        lcd.backlight()
#        lcd.delayMiliseconds(500)
#        lcd.noBacklight()
#        lcd.delayMiliseconds(500)
#
#    lcd.backlight()
#    lcd.clear()
#    lcd.delayMiliseconds(500)


board = telemetrix.Telemetrix()
try:
    pcf8574(board)
except KeyboardInterrupt:
    board.shutdown()
    sys.exit(0)



import sys
import time

from telemetrix import telemetrix

"""
Setup a pin for digital output 
and toggle the pin 5 times.
"""

# some globals
DIGITAL_PIN = 14  # the board LED

def the_callback(data):
    print(data)


# Create a Telemetrix instance.
board = telemetrix.Telemetrix()

# Set the DIGITAL_PIN as an output pin
board.set_pin_mode_dht(DIGITAL_PIN,  the_callback)

while True:
    time.sleep(5)
    print('exp')



"""
 Copyright (c) 2023 Alan Yorinks All rights reserved.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
 Version 3 as published by the Free Software Foundation; either
 or (at your option) any later version.
 This library is distributed in the hope that it will be useful,f
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License for more details.

 You should have received a copy of the GNU AFFERO GENERAL PUBLIC LICENSE
 along with this library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

 DHT support courtesy of Martyn Wheeler
 Based on the DHTNew library - https://github.com/RobTillaart/DHTNew
"""

import sys
import time

from telemetrix import telemetrix

"""
Simulate a rotary encoder
"""

"""
Setup pins for digital inputs and monitor their changes changes
"""

# Set up a pin for analog input and monitor its changes
CLOCK_WISE_PIN = 12  # arduino pin number
ANIT_CLOCK_WISE_PIN = 13

# some globals
astate = None
last_state = None
update = False
counter = 0


def the_callback(data):
    """
    A callback function to report data changes

    :param data: [report type, pin, current reported value, pin_mode, timestamp]
    """
    global astate, last_state, update, counter
    the_pin = data[1]
    the_pin_state = data[2]

    if the_pin == CLOCK_WISE_PIN:
        astate = the_pin_state
        if astate != last_state:
            last_state = astate
            update = True
    else:
        if update:
            if the_pin_state != astate:
                counter = counter + 1
            else:
                counter = counter - 1
            update = False
    print(f'astate: {astate}   last_state: {last_state}   update: {update}    counter: '
          f'{counter}')


def rotary_simulator(my_board, clockwise_pin, anti_clockwise_pin):
    """
     This function establishes the pins as
     digital inputs. Any changes on the pins will
     be reported through the call back function.

     :param my_board: a telemetrix instance
     :param clockwise_pin: Arduino pin number
     :param anti_clockwise_pin : Arduino pin number
     """

    # set the pin mode
    my_board.set_pin_mode_digital_input(clockwise_pin, the_callback)
    my_board.set_pin_mode_digital_input(anti_clockwise_pin, the_callback)

    try:
        while True:
            time.sleep(.0001)
    except KeyboardInterrupt:
        board.shutdown()
        sys.exit(0)


board = telemetrix.Telemetrix(arduino_wait=4)

try:
    rotary_simulator(board, CLOCK_WISE_PIN, ANIT_CLOCK_WISE_PIN)
except KeyboardInterrupt:
    board.shutdown()
    sys.exit(0)



"""
 Copyright (c) 2020 Alan Yorinks All rights reserved.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
 Version 3 as published by the Free Software Foundation; either
 or (at your option) any later version.
 This library is distributed in the hope that it will be useful,f
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License for more details.

 You should have received a copy of the GNU AFFERO GENERAL PUBLIC LICENSE
 along with this library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

 DHT support courtesy of Martyn Wheeler
 Based on the DHTNew library - https://github.com/RobTillaart/DHTNew
"""

import sys
import time

from telemetrix import telemetrix

"""
Setup a pin for output and fade its intensity
"""

# some globals
# make sure to select a PWM pin
DIGITAL_PIN = 6
SERVO_PIN = 5

# Create a Telemetrix instance.
board = telemetrix.Telemetrix()

# Set the DIGITAL_PIN as an output pin
board.set_pin_mode_analog_output(DIGITAL_PIN)
board.set_pin_mode_analog_output(9)
board.set_pin_mode_analog_output(10)
board.set_pin_mode_analog_output(11)

# board.set_pin_mode_analog_output(DIGITAL_PIN)
board.set_pin_mode_servo(SERVO_PIN, 100, 3000)
time.sleep(.2)
board.servo_write(SERVO_PIN, 90)
time.sleep(1)
board.servo_write(SERVO_PIN, 0)
time.sleep(1)

# When hitting control-c to end the program
# in this loop, we are likely to get a KeyboardInterrupt
# exception. Catch the exception and exit gracefully.

try:

    time.sleep(.2)
    board.servo_write(SERVO_PIN, 90)
    time.sleep(1)
    while True:
        print('Fading up...')
        board.servo_write(SERVO_PIN, 90)

        for i in range(255):
            board.analog_write(DIGITAL_PIN, i)
            board.analog_write(9, i)
            board.analog_write(10, i)
            board.analog_write(11, i)

            time.sleep(.005)
        print('Fading down...')
        board.servo_write(SERVO_PIN, 00)

        for i in range(255, -1, -1):
            board.analog_write(DIGITAL_PIN, i)
            board.analog_write(9, i)
            board.analog_write(10, i)
            board.analog_write(11, i)
            time.sleep(.005)

        time.sleep(3)


except KeyboardInterrupt:
    board.shutdown()
    sys.exit(0)



"""
 Copyright (c) 2020 Alan Yorinks All rights reserved.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
 Version 3 as published by the Free Software Foundation; either
 or (at your option) any later version.
 This library is distributed in the hope that it will be useful,f
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License for more details.

 You should have received a copy of the GNU AFFERO GENERAL PUBLIC LICENSE
 along with this library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

"""
import struct
import sys
import threading
import time
from collections import deque

import serial
# noinspection PyPackageRequirementscd
from serial.serialutil import SerialException
# noinspection PyPackageRequirements
from serial.tools import list_ports

from telemetrix.private_constants import PrivateConstants


# noinspection PyPep8,PyMethodMayBeStatic
class Telemetrix(threading.Thread):
    """
    This class exposes and implements the telemetrix API.
    It uses threading to accommodate concurrency.
    It includes the public API methods as well as
    a set of private methods.

    """

    # noinspection PyPep8,PyPep8,PyPep8
    def __init__(self, com_port=None, arduino_instance_id=1,
                 arduino_wait=4, sleep_tune=0.000001,
                 shutdown_on_exception=True):

        """

        :param com_port: e.g. COM3 or /dev/ttyACM0.
                         Only use if you wish to bypass auto com port
                         detection.

        :param arduino_instance_id: Match with the value installed on the
                                    arduino-telemetrix sketch.

        :param arduino_wait: Amount of time to wait for an Arduino to
                             fully reset itself.

        :param sleep_tune: A tuning parameter (typically not changed by user)

        :param shutdown_on_exception: call shutdown before raising
                                      a RunTimeError exception, or
                                      receiving a KeyboardInterrupt exception
        """

        # initialize threading parent
        threading.Thread.__init__(self)

        # create the threads and set them as daemons so
        # that they stop when the program is closed

        # create a thread to interpret received serial data
        self.the_reporter_thread = threading.Thread(target=self._reporter)
        self.the_reporter_thread.daemon = True

        self.the_data_receive_thread = threading.Thread(target=self._serial_receiver)

        self.the_data_receive_thread.daemon = True

        # flag to allow the reporter and receive threads to run.
        self.run_event = threading.Event()

        # check to make sure that Python interpreter is version 3.7 or greater
        python_version = sys.version_info
        if python_version[0] >= 3:
            if python_version[1] >= 7:
                pass
            else:
                raise RuntimeError("ERROR: Python 3.7 or greater is "
                                   "required for use of this program.")

        # save input parameters as instance variables
        self.com_port = com_port
        self.arduino_instance_id = arduino_instance_id
        self.arduino_wait = arduino_wait
        self.sleep_tune = sleep_tune
        self.shutdown_on_exception = shutdown_on_exception

        # create a deque to receive and process data from the arduino
        self.the_deque = deque()

        # The report_dispatch dictionary is used to process
        # incoming report messages by looking up the report message
        # and executing its associated processing method.

        self.report_dispatch = {}

        # To add a command to the command dispatch table, append here.
        self.report_dispatch.update({PrivateConstants.LOOP_COMMAND: self._report_loop_data})
        self.report_dispatch.update({PrivateConstants.DEBUG_PRINT: self._report_debug_data})
        self.report_dispatch.update({PrivateConstants.DIGITAL_REPORT: self._digital_message})
        self.report_dispatch.update({PrivateConstants.ANALOG_REPORT: self._analog_message})
        self.report_dispatch.update({PrivateConstants.FIRMWARE_REPORT: self._firmware_message})
        self.report_dispatch.update({PrivateConstants.I_AM_HERE_REPORT: self._i_am_here})
        self.report_dispatch.update({PrivateConstants.SERVO_UNAVAILABLE: self._servo_unavailable})
        self.report_dispatch.update({PrivateConstants.I2C_READ_REPORT: self._i2c_read_report})
        self.report_dispatch.update({PrivateConstants.I2C_TOO_FEW_BYTES_RCVD: self._i2c_too_few})
        self.report_dispatch.update({PrivateConstants.I2C_TOO_MANY_BYTES_RCVD: self._i2c_too_many})
        self.report_dispatch.update({PrivateConstants.SONAR_DISTANCE: self._sonar_distance_report})
        self.report_dispatch.update({PrivateConstants.DHT_REPORT: self._dht_report})

        # dictionaries to store the callbacks for each pin
        self.analog_callbacks = {}

        self.digital_callbacks = {}

        self.i2c_callback = None
        self.i2c_callback2 = None

        self.i2c_1_active = False
        self.i2c_2_active = False

        # the trigger pin will be the key to retrieve
        # the callback for a specific HC-SR04
        self.sonar_callbacks = {}

        self.sonar_count = 0

        self.dht_callbacks = {}

        self.dht_count = 0

        # serial port in use
        self.serial_port = None

        # flag to indicate we are in shutdown mode
        self.shutdown_flag = False

        # debug loopback callback method
        self.loop_back_callback = None

        # flag to indicate the start of a new report
        # self.new_report_start = True

        # firmware version to be stored here
        self.firmware_version = []

        # reported arduino instance id
        self.reported_arduino_id = []

        # flag to indicate if i2c was previously enabled
        self.i2c_enabled = False

        self.the_reporter_thread.start()
        self.the_data_receive_thread.start()

        print(f"Telemetrix:  Version {PrivateConstants.TELEMETRIX_VERSION}\n\n"
              f"Copyright (c) 2020 Alan Yorinks All Rights Reserved.\n")

        if not self.com_port:
            # user did not specify a com_port
            try:
                self._find_arduino()
            except KeyboardInterrupt:
                if self.shutdown_on_exception:
                    self.shutdown()
        else:
            # com_port specified - set com_port and baud rate
            try:
                self._manual_open()
            except KeyboardInterrupt:
                if self.shutdown_on_exception:
                    self.shutdown()

        if self.serial_port:
            print(f"Arduino compatible device found and connected to {self.serial_port.port}")

            self.serial_port.reset_input_buffer()
            self.serial_port.reset_output_buffer()

        # no com_port found - raise a runtime exception
        else:
            if self.shutdown_on_exception:
                self.shutdown()
            raise RuntimeError('No Arduino Found or User Aborted Program')

        # allow the threads to run
        self._run_threads()
        print(f'Waiting for Arduino to reset')
        print(f'Reset Complete')

        print('Retrieving Arduino ID...')
        self._get_arduino_id()
        if self.reported_arduino_id != self.arduino_instance_id:
            if self.shutdown_on_exception:
                self.shutdown()
            raise RuntimeError(f'Incorrect Arduino ID: {self.reported_arduino_id}')
        print('Valid Arduino ID Found.')
        # get arduino firmware version and print it
        print('\nRetrieving Telemetrix4Arduino firmware ID...')
        self._get_firmware_version()
        if not self.firmware_version:
            if self.shutdown_on_exception:
                self.shutdown()
            raise RuntimeError(f'Telemetrix4Arduino firmware version')

        else:
            print(f'Telemetrix4Arduino firmware version: {self.firmware_version[0]}.'
                  f'{self.firmware_version[1]}')

    def _find_arduino(self):
        """
        This method will search all potential serial ports for an Arduino
        containing a sketch that has a matching arduino_instance_id as
        specified in the input parameters of this class.

        This is used explicitly with the Telemetrix4Arduino sketch.
        """

        # a list of serial ports to be checked
        serial_ports = []

        print('Opening all potential serial ports...')
        the_ports_list = list_ports.comports()
        for port in the_ports_list:
            if port.pid is None:
                continue
            try:
                self.serial_port = serial.Serial(port.device, 115200,
                                                 timeout=1, writeTimeout=0)
            except SerialException:
                continue
            # create a list of serial ports that we opened
            serial_ports.append(self.serial_port)

            # display to the user
            print('\t' + port.device)

            # clear out any possible data in the input buffer
        # wait for arduino to reset
        print(f'\nWaiting {self.arduino_wait} seconds(arduino_wait) for Arduino devices to '
              'reset...')
        time.sleep(self.arduino_wait)

        # check for correct arduino device
        self.serial_port.reset_input_buffer()
        self.serial_port.reset_output_buffer()

    def _manual_open(self):
        """
        Com port was specified by the user - try to open up that port

        """
        # if port is not found, a serial exception will be thrown
        try:
            print(f'Opening {self.com_port}...')
            self.serial_port = serial.Serial(self.com_port, 115200,
                                             timeout=1, writeTimeout=0)

            print(f'\nWaiting {self.arduino_wait} seconds(arduino_wait) for Arduino devices to '
                  'reset...')
            self._run_threads()
            time.sleep(self.arduino_wait)

            self._get_arduino_id()

            if self.reported_arduino_id != self.arduino_instance_id:
                if self.shutdown_on_exception:
                    self.shutdown()
                raise RuntimeError(f'Incorrect Arduino ID: {self.reported_arduino_id}')
            print('Valid Arduino ID Found.')
            # get arduino firmware version and print it
            print('\nRetrieving Telemetrix4Arduino firmware ID...')
            self._get_firmware_version()

            if not self.firmware_version:
                if self.shutdown_on_exception:
                    self.shutdown()
                raise RuntimeError(f'Telemetrix4Arduino Sketch Firmware Version Not Found')

            else:
                print(f'Telemetrix4Arduino firmware version: {self.firmware_version[0]}.'
                      f'{self.firmware_version[1]}')
        except KeyboardInterrupt:
            if self.shutdown_on_exception:
                self.shutdown()
            raise RuntimeError('User Hit Control-C')

    def analog_write(self, pin, value):
        """
        Set the specified pin to the specified value.

        :param pin: arduino pin number

        :param value: pin value (0-255)

        """
        command = [PrivateConstants.ANALOG_WRITE, pin, value]
        self._send_command(command)

    def digital_write(self, pin, value):
        """
        Set the specified pin to the specified value.

        :param pin: arduino pin number

        :param value: pin value (1 or 0)

        """

        command = [PrivateConstants.DIGITAL_WRITE, pin, value]
        self._send_command(command)

    def disable_all_reporting(self):
        """
        Disable reporting for all digital and analog input pins
        """
        command = [PrivateConstants.MODIFY_REPORTING, PrivateConstants.REPORTING_DISABLE_ALL, 0]
        self._send_command(command)

    def disable_analog_reporting(self, pin):
        """
        Disables analog reporting for a single analog pin.

        :param pin: Analog pin number. For example for A0, the number is 0.

        """
        command = [PrivateConstants.MODIFY_REPORTING, PrivateConstants.REPORTING_ANALOG_DISABLE, pin]
        self._send_command(command)

    def disable_digital_reporting(self, pin):
        """
        Disables digital reporting for a single digital input.

        :param pin: Pin number.

        """
        command = [PrivateConstants.MODIFY_REPORTING, PrivateConstants.REPORTING_DIGITAL_DISABLE, pin]
        self._send_command(command)

    def enable_analog_reporting(self, pin):
        """
        Enables analog reporting for the specified pin.

        :param pin: Analog pin number. For example for A0, the number is 0.


        """
        command = [PrivateConstants.MODIFY_REPORTING, PrivateConstants.REPORTING_ANALOG_ENABLE, pin]
        self._send_command(command)

    def enable_digital_reporting(self, pin):
        """
        Enable reporting on the specified digital pin.

        :param pin: Pin number.
        """

        command = [PrivateConstants.MODIFY_REPORTING, PrivateConstants.REPORTING_DIGITAL_ENABLE, pin]
        self._send_command(command)

    def _get_arduino_id(self):
        """
        Retrieve arduino-telemetrix arduino id

        """
        command = [PrivateConstants.ARE_U_THERE]
        self._send_command(command)
        # provide time for the reply
        time.sleep(.1)

    def _get_firmware_version(self):
        """
        This method retrieves the
        arduino-telemetrix firmware version

        """
        command = [PrivateConstants.GET_FIRMWARE_VERSION]
        self._send_command(command)
        # provide time for the reply
        time.sleep(.1)

    def i2c_read(self, address, register, number_of_bytes,
                 callback=None, i2c_port=0):
        """
        Read the specified number of bytes from the specified register for
        the i2c device.


        :param address: i2c device address

        :param register: i2c register (or None if no register selection is needed)

        :param number_of_bytes: number of bytes to be read

        :param callback: Required callback function to report i2c data as a
                   result of read command

       :param i2c_port: 0 = default, 1 = secondary


        callback returns a data list:
        [I2C_READ_REPORT, address, register, count of data bytes, data bytes, time-stamp]

        """

        self._i2c_read_request(address, register, number_of_bytes,
                               callback=callback, i2c_port=i2c_port)

    def i2c_read_restart_transmission(self, address, register,
                                      number_of_bytes,
                                      callback=None, i2c_port=0):
        """
        Read the specified number of bytes from the specified register for
        the i2c device. This restarts the transmission after the read. It is
        required for some i2c devices such as the MMA8452Q accelerometer.


        :param address: i2c device address

        :param register: i2c register (or None if no register
                                                    selection is needed)

        :param number_of_bytes: number of bytes to be read

        :param callback: Required callback function to report i2c data as a
                   result of read command

       :param i2c_port: 0 = default 1 = secondary


        callback returns a data list:

        [I2C_READ_REPORT, address, register, count of data bytes, data bytes, time-stamp]

        """

        self._i2c_read_request(address, register, number_of_bytes, stop_transmission=False,
                               callback=callback, i2c_port=i2c_port)

    def _i2c_read_request(self, address, register, number_of_bytes,
                          stop_transmission=True, callback=None, i2c_port=0):
        """
        This method requests the read of an i2c device. Results are retrieved
        via callback.

        :param address: i2c device address

        :param register: register number (or None if no register selection is needed)

        :param number_of_bytes: number of bytes expected to be returned

        :param stop_transmission: stop transmission after read

        :param callback: Required callback function to report i2c data as a
                   result of read command.

        """
        if not i2c_port:
            if not self.i2c_1_active:
                if self.shutdown_on_exception:
                    self.shutdown()
                raise RuntimeError('I2C Read: set_pin_mode i2c never called for i2c port 1.')

        if i2c_port:
            if not self.i2c_2_active:
                if self.shutdown_on_exception:
                    self.shutdown()
                raise RuntimeError('I2C Read: set_pin_mode i2c never called for i2c port 2.')

        if not callback:
            if self.shutdown_on_exception:
                self.shutdown()
            raise RuntimeError('I2C Read: A callback function must be specified.')

        if not i2c_port:
            self.i2c_callback = callback
        else:
            self.i2c_callback2 = callback

        if not register:
            register = 0

        # message contains:
        # 1. address
        # 2. register
        # 3. number of bytes
        # 4. restart_transmission - True or False
        # 5. i2c port

        command = [PrivateConstants.I2C_READ, address, register, number_of_bytes,
                   stop_transmission, i2c_port]
        self._send_command(command)

    def i2c_write(self, address, args, i2c_port=0):
        """
        Write data to an i2c device.

        :param address: i2c device address

        :param i2c_port: 0= port 1, 1 = port 2

        :param args: A variable number of bytes to be sent to the device
                     passed in as a list

        """
        if not i2c_port:
            if not self.i2c_1_active:
                if self.shutdown_on_exception:
                    self.shutdown()
                raise RuntimeError('I2C Write: set_pin_mode i2c never called for i2c port 1.')

        if i2c_port:
            if not self.i2c_2_active:
                if self.shutdown_on_exception:
                    self.shutdown()
                raise RuntimeError('I2C Write: set_pin_mode i2c never called for i2c port 2.')

        command = [PrivateConstants.I2C_WRITE, len(args), address, i2c_port]

        for item in args:
            command.append(item)

        self._send_command(command)

    def loop_back(self, start_character, callback=None):
        """
        This is a debugging method to send a character to the
        Arduino device, and have the device loop it back.

        :param start_character: The character to loop back. It should be
                                an integer.

        :param callback: Looped back character will appear in the callback method

        """
        command = [PrivateConstants.LOOP_COMMAND, ord(start_character)]
        self.loop_back_callback = callback
        self._send_command(command)

    def set_analog_scan_interval(self, interval):
        """
        Set the analog scanning interval.

        :param interval: value of 0 - 255 - milliseconds
        """

        if 0 <= interval <= 255:
            command = [PrivateConstants.SET_ANALOG_SCANNING_INTERVAL, interval]
            self._send_command(command)
        else:
            if self.shutdown_on_exception:
                self.shutdown()
            raise RuntimeError('Analog interval must be between 0 and 255')

    def set_pin_mode_analog_output(self, pin_number):
        """
        Set a pin as a pwm (analog output) pin.

        :param pin_number:arduino pin number

        """
        self._set_pin_mode(pin_number, PrivateConstants.AT_OUTPUT)

    def set_pin_mode_analog_input(self, pin_number, differential=0, callback=None):
        """
        Set a pin as an analog input.

        :param pin_number: arduino pin number

        :param differential: difference in previous to current value before
                             report will be generated

        :param callback: callback function


        callback returns a data list:

        [pin_type, pin_number, pin_value, raw_time_stamp]

        The pin_type for analog input pins = 2

        """
        self._set_pin_mode(pin_number, PrivateConstants.AT_ANALOG, differential,
                           callback)

    def set_pin_mode_digital_input(self, pin_number, callback=None):
        """
        Set a pin as a digital input.

        :param pin_number: arduino pin number

        :param callback: callback function


        callback returns a data list:

        [pin_type, pin_number, pin_value, raw_time_stamp]

        The pin_type for digital input pins = 0

        """
        self._set_pin_mode(pin_number, PrivateConstants.AT_INPUT, callback=callback)

    def set_pin_mode_digital_input_pullup(self, pin_number, callback=None):
        """
        Set a pin as a digital input with pullup enabled.

        :param pin_number: arduino pin number

        :param callback: callback function


        callback returns a data list:

        [pin_type, pin_number, pin_value, raw_time_stamp]

        The pin_type for digital input pins with pullups enabled = 11

        """
        self._set_pin_mode(pin_number, PrivateConstants.AT_INPUT_PULLUP, callback)

    def set_pin_mode_digital_output(self, pin_number):
        """
        Set a pin as a digital output pin.

        :param pin_number: arduino pin number
        """

        self._set_pin_mode(pin_number, PrivateConstants.AT_OUTPUT)

    def set_pin_mode_i2c(self, i2c_port=0):
        """
        Establish the standard Arduino i2c pins for i2c utilization.

        :param i2c_port: 0 = i2c1, 1 = i2c2

        NOTES: 1. THIS METHOD MUST BE CALLED BEFORE ANY I2C REQUEST IS MADE
               2. Callbacks are set within the individual i2c read methods of this
              API.

              See i2c_read, or i2c_read_restart_transmission.

        """
        # test for i2c port 2
        if i2c_port:
            # if not previously activated set it to activated
            # and the send a begin message for this port
            if not self.i2c_2_active:
                self.i2c_2_active = True
            else:
                return
        # port 1
        else:
            if not self.i2c_1_active:
                self.i2c_1_active = True
            else:
                return

        command = [PrivateConstants.I2C_BEGIN, i2c_port]
        self._send_command(command)

    def set_pin_mode_dht(self, pin, callback=None):
        """

        :param pin: connection pin

        :param callback: callback function

        Error Callback: [Callback 0=DHT REPORT, DHT_ERROR=0, PIN, Error Number, Time]

        Valid Data Callback: Callback 0=DHT REPORT, DHT_DATA=1, PIN, Humidity, Temperature Time]

        """

        if not callback:
            if self.shutdown_on_exception:
                self.shutdown()
            raise RuntimeError('set_pin_mode_dht: A Callback must be specified')

        if self.dht_count < PrivateConstants.MAX_DHTS - 1:
            self.dht_callbacks[pin] = callback
            self.dht_count += 1

            command = [PrivateConstants.DHT_NEW, pin]
            self._send_command(command)
        else:
            if self.shutdown_on_exception:
                self.shutdown()
            raise RuntimeError(f'Maximum Number Of DHTs Exceeded - set_pin_mode_dht fails for pin {pin}')

    # noinspection PyRedundantParentheses
    def set_pin_mode_servo(self, pin_number, min_pulse=544, max_pulse=2400):
        """

        Attach a pin to a servo motor

        :param pin_number: pin

        :param min_pulse: minimum pulse width

        :param max_pulse: maximum pulse width

        """
        minv = (min_pulse).to_bytes(2, byteorder="big")
        maxv = (max_pulse).to_bytes(2, byteorder="big")

        command = [PrivateConstants.SERVO_ATTACH, pin_number,
                   minv[0], minv[1], maxv[0], maxv[1]]
        self._send_command(command)

    def set_pin_mode_sonar(self, trigger_pin, echo_pin,
                           callback=None):
        """

        :param trigger_pin:

        :param echo_pin:

        :param callback: callback

        callback data: [PrivateConstants.SONAR_DISTANCE, trigger_pin, distance_value, time_stamp]

        """

        if not callback:
            if self.shutdown_on_exception:
                self.shutdown()
            raise RuntimeError('set_pin_mode_sonar: A Callback must be specified')

        if self.sonar_count < PrivateConstants.MAX_SONARS - 1:
            self.sonar_callbacks[trigger_pin] = callback
            self.sonar_count += 1

            command = [PrivateConstants.SONAR_NEW, trigger_pin, echo_pin]
            self._send_command(command)
        else:
            if self.shutdown_on_exception:
                self.shutdown()
            raise RuntimeError(f'Maximum Number Of Sonars Exceeded - set_pin_mode_sonar fails for pin {trigger_pin}')

    def servo_write(self, pin_number, angle):
        """

        Set a servo attached to a pin to a given angle.

        :param pin_number: pin

        :param angle: angle (0-180)

        """
        command = [PrivateConstants.SERVO_WRITE, pin_number, angle]
        self._send_command(command)

    def servo_detach(self, pin_number):
        """
        Detach a servo for reuse

        :param pin_number: attached pin

        """
        command = [PrivateConstants.SERVO_DETACH, pin_number]
        self._send_command(command)

    def _set_pin_mode(self, pin_number, pin_state, differential=0, callback=None):

        """
        A private method to set the various pin modes.

        :param pin_number: arduino pin number

        :param pin_state: INPUT/OUTPUT/ANALOG/PWM/PULLUP
                         For SERVO use: set_pin_mode_servo
                         For DHT   use: set_pin_mode_dht

        :param differential: for analog inputs - threshold
                             value to be achieved for report to
                             be generated

        :param callback: A reference to a call back function to be
                         called when pin data value changes

        """
        if callback:
            if pin_state == PrivateConstants.AT_INPUT:
                self.digital_callbacks[pin_number] = callback
            elif pin_state == PrivateConstants.AT_INPUT_PULLUP:
                self.digital_callbacks[pin_number] = callback
            elif pin_state == PrivateConstants.AT_ANALOG:
                self.analog_callbacks[pin_number] = callback
            else:
                print('{} {}'.format('set_pin_mode: callback ignored for '
                                     'pin state:', pin_state))

        if pin_state == PrivateConstants.AT_INPUT:
            command = [PrivateConstants.SET_PIN_MODE, pin_number, PrivateConstants.AT_INPUT, 1]

        elif pin_state == PrivateConstants.AT_INPUT_PULLUP:
            command = [PrivateConstants.SET_PIN_MODE, pin_number, PrivateConstants.AT_INPUT_PULLUP, 1]

        elif pin_state == PrivateConstants.AT_OUTPUT:
            command = [PrivateConstants.SET_PIN_MODE, pin_number, PrivateConstants.AT_OUTPUT]

        elif pin_state == PrivateConstants.AT_ANALOG:
            command = [PrivateConstants.SET_PIN_MODE, pin_number, PrivateConstants.AT_ANALOG,
                       differential >> 8, differential & 0xff, 1]
        else:
            if self.shutdown_on_exception:
                self.shutdown()
            raise RuntimeError('Unknown pin state')

        if command:
            self._send_command(command)

    def shutdown(self):
        """
        This method attempts an orderly shutdown
        If any exceptions are thrown, they are ignored.
        """
        self.shutdown_flag = True

        self._stop_threads()

        try:
            self.disable_all_reporting()
            command = [PrivateConstants.STOP_ALL_REPORTS]
            self._send_command(command)
            time.sleep(.5)

            self.serial_port.reset_input_buffer()
            self.serial_port.close()

        except (RuntimeError, SerialException, OSError):
            # ignore error on shutdown
            pass

    '''
    report message handlers
    '''

    def _analog_message(self, data):
        """
        This is a private message handler method.
        It is a message handler for analog messages.

        :param data: message data

        """
        pin = data[0]
        value = (data[1] << 8) + data[2]
        # set the current value in the pin structure
        time_stamp = time.time()
        # self.digital_pins[pin].event_time = time_stamp
        if self.analog_callbacks[pin]:
            message = [PrivateConstants.ANALOG_REPORT, pin, value, time_stamp]
            self.analog_callbacks[pin](message)

    def _dht_report(self, data):
        """
        This is the dht report handler method.

        :param data:            data[0] = report sub type - DHT_DATA or DHT_ERROR

                                data[1] = pin number

                                data[2] = humidity high order byte or error value if DHT_ERROR

                                data[3] = humidity byte 2

                                data[4] = humidity byte 3

                                data[5] = humidity byte 4

                                data[6] = temperature high order byte for data

                                data[7] = temperature byte 2

                                data[8] = temperature byte 3

                                data[9] = temperature byte 4
        """

        if data[0]:  # DHT_ERROR
            # error report
            # data[0] = report sub type, data[1] = pin, data[2] = error message
            if self.dht_callbacks[data[1]]:
                # Callback 0=DHT REPORT, DHT_ERROR=0, PIN, Error Number, Time
                message = [PrivateConstants.DHT_REPORT, data[0], data[1], data[2], time.time()]
                self.dht_callbacks[data[1]](message)
        else:
            # got valid data DHT_DATA
            f_humidity = bytearray(data[2:6])
            f_temperature = bytearray(data[6:])
            message = [PrivateConstants.DHT_REPORT, data[0], data[1],
                       (struct.unpack('<f', f_humidity))[0],
                       (struct.unpack('<f', f_temperature))[0],
                       time.time()]

            self.dht_callbacks[data[1]](message)

    def _digital_message(self, data):
        """
        This is a private message handler method.
        It is a message handler for Digital Messages.

        :param data: digital message

        """
        pin = data[0]
        value = data[1]

        time_stamp = time.time()
        if self.digital_callbacks[pin]:
            message = [PrivateConstants.DIGITAL_REPORT, pin, value, time_stamp]
            self.digital_callbacks[pin](message)

    def _firmware_message(self, data):
        """
        Telemetrix4Arduino firmware version message
        :param data: data[0] = major number, data[1] = minor number
        """

        self.firmware_version = [data[0], data[1]]

    def _i2c_read_report(self, data):
        """
        Execute callback for i2c reads.

        :param data: [I2C_READ_REPORT, i2c_port, number of bytes read, address, register, bytes read..., time-stamp]
        """

        # we receive [# data bytes, address, register, data bytes]
        # number of bytes of data returned

        # data[0] = number of bytes
        # data[1] = i2c_port
        # data[2] = number of bytes returned
        # data[3] = address
        # data[4] = register
        # data[5] ... all the data bytes

        cb_list = [PrivateConstants.I2C_READ_REPORT, data[0], data[1]] + data[2:]
        cb_list.append(time.time())

        if cb_list[1]:
            self.i2c_callback2(cb_list)
        else:
            self.i2c_callback(cb_list)

    def _i2c_too_few(self, data):
        """
        I2c reports too few bytes received

        :param data: data[0] = device address
        """
        if self.shutdown_on_exception:
            self.shutdown()
        raise RuntimeError(f'i2c too few bytes received from i2c port {data[0]} i2c address {data[1]}')

    def _i2c_too_many(self, data):
        """
        I2c reports too few bytes received

        :param data: data[0] = device address
        """
        if self.shutdown_on_exception:
            self.shutdown()
        raise RuntimeError(f'i2c too many bytes received from i2c port {data[0]} i2c address {data[1]}')

    def _i_am_here(self, data):
        """
        Reply to are_u_there message
        :param data: arduino id
        """
        self.reported_arduino_id = data[0]

    def _report_debug_data(self, data):
        """
        Print debug data sent from Arduino
        :param data: data[0] is a byte followed by 2
                     bytes that comprise an integer
        :return:
        """
        value = (data[1] << 8) + data[2]
        print(f'DEBUG ID: {data[0]} Value: {value}')

    def _report_loop_data(self, data):
        """
        Print data that was looped back
        :param data: byte of loop back data
        :return:
        """
        if self.loop_back_callback:
            self.loop_back_callback(data)

    def _send_command(self, command):
        """
        This is a private utility method.


        :param command:  command data in the form of a list

        :returns: number of bytes sent
        """
        # the length of the list is added at the head
        command.insert(0, len(command))
        # print(command)
        send_message = bytes(command)
        try:
            self.serial_port.write(send_message)
            # print(send_message)
        except SerialException:
            if self.shutdown_on_exception:
                self.shutdown()
            raise RuntimeError('write fail in _send_command')

    def _servo_unavailable(self, report):
        """
        Message if no servos are available for use.
        :param report: pin number
        """
        if self.shutdown_on_exception:
            self.shutdown()
        raise RuntimeError(f'Servo Attach For Pin {report[0]} Failed: No Available Servos')

    def _sonar_distance_report(self, report):
        """

        :param report: data[0] = trigger pin, data[1] and data[2] = distance

        callback report format: [PrivateConstants.SONAR_DISTANCE, trigger_pin, distance_value, time_stamp]
        """

        # get callback from pin number
        cb = self.sonar_callbacks[report[0]]

        # build report data
        cb_list = [PrivateConstants.SONAR_DISTANCE, report[0],
                   ((report[1] << 8) + report[2]), time.time()]

        cb(cb_list)

    def _run_threads(self):
        self.run_event.set()

    def _is_running(self):
        return self.run_event.is_set()

    def _stop_threads(self):
        self.run_event.clear()

    def _reporter(self):
        """
        This is the reporter thread. It continuously pulls data from
        the deque. When a full message is detected, that message is
        processed.
        """
        self.run_event.wait()

        while self._is_running() and not self.shutdown_flag:
            if len(self.the_deque):
                # response_data will be populated with the received data for the report
                response_data = []
                packet_length = self.the_deque.popleft()
                if packet_length:
                    # get all the data for the report and place it into response_data
                    for i in range(packet_length):
                        while not len(self.the_deque):
                            time.sleep(self.sleep_tune)
                        data = self.the_deque.popleft()
                        response_data.append(data)

                    # print(response_data)

                    # get the report type and look up its dispatch method
                    # here we pop the report type off of response_data
                    report_type = response_data.pop(0)
                    # print(report_type)

                    # retrieve the report handler from the dispatch table
                    dispatch_entry = self.report_dispatch.get(report_type)

                    # if there is additional data for the report,
                    # it will be contained in response_data
                    # noinspection PyArgumentList
                    dispatch_entry(response_data)
                    continue

                else:
                    if self.shutdown_on_exception:
                        self.shutdown()
                    raise RuntimeError('A report with a packet length of zero was received.')
            else:
                time.sleep(self.sleep_tune)

    def _serial_receiver(self):
        """
        Thread to continuously check for incoming data.
        When a byte comes in, place it onto the deque.
        """
        self.run_event.wait()

        while self._is_running() and not self.shutdown_flag:
            # we can get an OSError: [Errno9] Bad file descriptor when shutting down
            # just ignore it
            try:
                if self.serial_port.inWaiting():
                    c = self.serial_port.read()
                    self.the_deque.append(ord(c))
                    # print(ord(c))
                else:
                    time.sleep(self.sleep_tune)
                    # continue
            except OSError:
                pass
